/***************************************************************************
 *   Copyright (C) 2005-2013 LpzRobots development team                    *
 *   Authors:                                                              *
 *    Simón Smith <artificialsimon at ed dot ac dot uk>                    *
 *    Georg Martius  <georg dot martius at web dot de>                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include "basiccontroller.h"

#include <assert.h>
#include <math.h>
using namespace std;
using namespace matrix;

BasicController::BasicController(const std::string& name)
  : AbstractController(name, "1.0") {
  initialised=false;
  // add threshold parameter to configurable parameters, setable on console
  addParameterDef("threshold",  &threshold, 0.18, 0, 1, "threshold for IR-sensor");
  addParameterDef("use_if",     &use_if, true, "Use IF construct? Alternative: additive scheme");
  addParameterDef("heaviside",  &heaviside, false, "Use Heaviside? Alternative: sigmoidal");
  addParameterDef("slope_sigm", &slope_sigm, 10.0, 0.1, 100.0, "Slope of sigmoidal");
  addParameterDef("use_vector", &use_vector, false, "Use Vector? Only used, when use_if is false");
  addParameterDef("base_speed", &base_speed, 1.0, 0, 5.0, "Base Speed for Vector Controller");
  
  // Adding one more parameter sometimes causes unresolved runtime errors 
  // Therefore 'slope_tanh' was excluded from runtime parameters
  // Instead it was defined in the 'stepNoLearning' function
  //addParameterDef("slope_tanh", &slope_tanh, 5.0, 0.5, 5000.0, "Slope of tanh");
  
  addInspectableValue("min_distance", &min_distance, "Minimal Distance to any Object");
}


double BasicController::get_max(const double *Array, int N){
    double max = 0.0;
    for (int i = 0; i < N; i++){
		if (Array[i] > max){
			max = Array[i];
		}
    }
    return max;
}


double BasicController::activation(double value, double bias=0.05){  
  if (heaviside){
    // Heaviside - Discrete two-valued logic
    if (value >= 0.0)
      return 1.0;
    else
      return 0.0;
  }
  else {
    // Sigmoid function - continuous fuzzy logic
    return 1.0/(1.0+exp(-slope_sigm*(value-bias)));
  }
}

void BasicController::stepNoLearning(const sensor* sensors, int number_sensors,
                                     motor* motors, int number_motors) {
  // Very simple controller, if IR sensors are higher than threshold then
  // the robot will turn oposite direction, or backward when to close
  // Sensors indices and motor indices can be accessed by the SIdx and MIdx functions
  // provided by the controller base class
  double threshold_sides = 3*threshold;
  double threshold_backward = 4*threshold;
  double threshold_front_sides = 2.5*threshold;
  double slope_tanh = 25.0;
  
  min_distance = 1 - get_max(sensors, nSensors);

// -------------------------- Controller Section -----------------------------
// If Controller
  if (use_if){
	if (sensors[SIdx("IR front left")] > threshold_backward &&
		sensors[SIdx("IR front right")] > threshold_backward) {
	  // move backward if something's in front of the robot
	  motors[0] = -.5; // right motor
	  motors[1] = -.5; // left motor
	}
	else if (sensors[SIdx("IR left")] > threshold_sides ||
		 sensors[SIdx("IR front left")] > threshold ||
		 sensors[SIdx("IR front left left")] > threshold_front_sides) {
	  // turn right if something's on the left side
	  motors[0] = -.25; // right motor
	  motors[1] = .75;  // left motor
	}
	else if (sensors[SIdx("IR right")] > threshold_sides ||
		 sensors[SIdx("IR front right")] > threshold ||
		 sensors[SIdx("IR front right right")] > threshold_front_sides) {
	  // turn left if something's on the right side
	  motors[0] = .75;  // right motor
	  motors[1] = -.25; // left motor
	}
	else {
	  // else move forward
	  motors[0] = 1.; // right motor
	  motors[1] = 1.; // left motor
	}
  }
// Theta-Heaviside and Sigmoid Controller
  else if (!use_vector) {
	// auxiliary variables; arguments of Heavisides
	double A1, A2, A3;             
	double B1, B21, B22, B31, B32;

	double C1, C2, C3;             // Conditions 1, 2, 3
								   // (results of algebraic representation of logical operators as Heavisides)
	double NOTC1, NOTC2, NOTC3;    // Negations of Conditions 1, 2, 3


	// ------------ Condition 1 ------------
	A1    = sensors[SIdx("IR front left")] - threshold_backward;
	B1    = sensors[SIdx("IR front right")] - threshold_backward;

	// C1 = A1 AND B1
	C1    = activation( A1) * activation( B1);
	NOTC1 = activation(-A1) * activation( B1) + 
			activation( A1) * activation(-B1) + 
			activation(-A1) * activation(-B1)

			
	// ------------ Condition 2 ------------
	A2    = sensors[SIdx("IR left")] - threshold_sides;
	B21   = sensors[SIdx("IR front left")] - threshold;
	B22   = sensors[SIdx("IR front left left")] - threshold_front_sides;

	// C2 = A2 OR B21 OR B22
	C2    = activation( A2) * activation(-B21) * activation(-B22) + 
			activation(-A2) * activation( B21) * activation(-B22) + 
			activation(-A2) * activation(-B21) * activation( B22) + 
			activation( A2) * activation( B21) * activation(-B22) +
			activation( A2) * activation(-B21) * activation( B22) + 
			activation(-A2) * activation( B21) * activation( B22) + 
			activation( A2) * activation( B21) * activation( B22);
	NOTC2 = activation(-A2) * activation(-B21) * activation(-B22);


	// ------------ Condition 3 ------------
	A3    = sensors[SIdx("IR right")] - threshold_sides;
	B31   = sensors[SIdx("IR front right")] - threshold;
	B32   = sensors[SIdx("IR front right right")] - threshold_front_sides;

	// C3 = A3 OR B31 OR B32
	C3    = activation( A3) * activation(-B31) * activation(-B32) +
			activation(-A3) * activation( B31) * activation(-B32) +
			activation(-A3) * activation(-B31) * activation( B32) + 
			activation( A3) * activation( B31) * activation(-B32) + 
			activation( A3) * activation(-B31) * activation( B32) + 
			activation(-A3) * activation( B31) * activation( B32) + 
			activation( A3) * activation( B31) * activation( B32);
	NOTC3 = activation(-A3) * activation(-B31) * activation(-B32);


	// ------------ Assign Motor values ------------
	//          backward +  turn right      +  turn left             + forward
	motors[0] = -0.5*C1  +  (-.25)*NOTC1*C2 +  ( .75)*NOTC1*NOTC2*C3 + 1.0*NOTC1*NOTC2*NOTC3;   // right motor
	motors[1] = -0.5*C1  +  ( .75)*NOTC1*C2 +  (-.25)*NOTC1*NOTC2*C3 + 1.0*NOTC1*NOTC2*NOTC3;   // left motor
  }
// Vector Controller
  else{
   double vecx =  	- 1.0  * (1-sensors[SIdx("IR left")]) 
					- 0.7  * (1-sensors[SIdx("IR front left left")])
					- 0.15 * (1-sensors[SIdx("IR front left")]) 
					- 0.3  * (1-sensors[SIdx("IR back left")])  
					+ 0.15 * (1-sensors[SIdx("IR front right")]) 
					+ 0.7  * (1-sensors[SIdx("IR front right right")]) 
					+ 1.0  * (1-sensors[SIdx("IR right")]) 
					+ 0.3  * (1-sensors[SIdx("IR back right")]) ;

   double vecy =    + 0.7  * (1-sensors[SIdx("IR front left left")]) 
					+ 0.98 * (1-sensors[SIdx("IR front left")]) 
					+ 0.98 * (1-sensors[SIdx("IR front right")]) 
					+ 0.7  * (1-sensors[SIdx("IR front right right")]) 
					- 1.68 * (1-sensors[SIdx("IR back left")]) 
					- 1.68 * (1-sensors[SIdx("IR back right")]) ;


   motors[0] = base_speed + vecy - vecx; // right motor
   motors[1] = base_speed + vecy + vecx; // left motor
  }
// ----------------------------------------------------------------------------
}

void BasicController::step(const sensor* sensors, int sensornumber,
                           motor* motors, int motornumber) {
  stepNoLearning(sensors,sensornumber, motors, motornumber);
}


void BasicController::init(int sensornumber, int motornumber, RandGen* randGen) {
  assert(motornumber >=2 && sensornumber >=6);
  // Set the number of sensors and motors
  nSensors = sensornumber;
  nMotors  = motornumber;
  initialised=true;
}


int BasicController::getSensorNumber() const {
  return nSensors;
}

int BasicController::getMotorNumber() const {
  return nMotors;
}

bool BasicController::store(FILE* f) const {
  //  S.store(f); // if S is a matrix::Matrix
  Configurable::print(f,0);
  return true;
}

bool BasicController::restore(FILE* f) {
  //  S.restore(f); // if S is a matrix::Matrix
  Configurable::parse(f);
  return true;
}
